# -*- coding: utf-8 -*-
# Generated by Django 1.10.4 on 2016-12-21 23:23
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("edc_metadata", "0003_auto_20161211_1005")]

    operations = [
        migrations.AlterField(
            model_name="crfmetadata",
            name="subject_identifier",
            field=models.CharField(max_length=50, verbose_name="Subject Identifier"),
        ),
        migrations.AlterField(
            model_name="requisitionmetadata",
            name="subject_identifier",
            field=models.CharField(max_length=50, verbose_name="Subject Identifier"),
        ),
    ]



from dateutil.relativedelta import relativedelta
from edc_visit_schedule import Crf, FormsCollection, Schedule, Visit, VisitSchedule
from edc_visit_schedule.constants import DAY1, MONTH1, MONTH3, MONTH6, WEEK2
from edc_visit_schedule.tests import DummyPanel

app_label = "edc_metadata"


class MockPanel(DummyPanel):
    """`requisition_model` is normally set when the lab profile
    is set up.
    """

    def __init__(self, name):
        super().__init__(requisition_model="edc_metadata.subjectrequisition", name=name)


crfs_prn = FormsCollection(
    Crf(show_order=100, model=f"{app_label}.prnone"),
    Crf(show_order=200, model=f"{app_label}.prntwo"),
)

crfs_missed = FormsCollection(
    Crf(show_order=1, model="edc_metadata.subjectvisitmissed", required=True),
)

crfs0 = FormsCollection(Crf(show_order=1, model=f"{app_label}.crftwo", required=False))
crfs1 = FormsCollection(Crf(show_order=1, model=f"{app_label}.crfone", required=True))
crfs2 = FormsCollection(Crf(show_order=1, model=f"{app_label}.crfone", required=False))
crfs3 = FormsCollection(Crf(show_order=1, model=f"{app_label}.crfone", required=False))
crfs4 = FormsCollection(Crf(show_order=1, model=f"{app_label}.crfone", required=False))


visit0 = Visit(
    code=DAY1,
    title="Week 1",
    timepoint=0,
    rbase=relativedelta(days=0),
    rlower=relativedelta(days=0),
    rupper=relativedelta(days=6),
    crfs=crfs0,
    crfs_prn=crfs_prn,
    allow_unscheduled=True,
    facility_name="5-day-clinic",
)

visit1 = Visit(
    code=WEEK2,
    title="Week 2",
    timepoint=1,
    rbase=relativedelta(days=7),
    rlower=relativedelta(days=0),
    rupper=relativedelta(days=6),
    crfs=crfs1,
    crfs_prn=crfs_prn,
    crfs_missed=crfs_missed,
    facility_name="5-day-clinic",
)

visit2 = Visit(
    code=MONTH1,
    title="Month 1",
    timepoint=2,
    rbase=relativedelta(month=1),
    rlower=relativedelta(days=0),
    rupper=relativedelta(days=6),
    crfs=crfs2,
    crfs_prn=crfs_prn,
    facility_name="5-day-clinic",
)

visit3 = Visit(
    code=MONTH3,
    title="Month 3",
    timepoint=3,
    rbase=relativedelta(month=3),
    rlower=relativedelta(days=0),
    rupper=relativedelta(days=6),
    crfs=crfs3,
    crfs_prn=crfs_prn,
    facility_name="5-day-clinic",
)

visit4 = Visit(
    code=MONTH6,
    title="Month 6",
    timepoint=4,
    rbase=relativedelta(month=6),
    rlower=relativedelta(days=0),
    rupper=relativedelta(days=6),
    crfs=crfs4,
    crfs_prn=crfs_prn,
    facility_name="5-day-clinic",
)

schedule = Schedule(
    name="schedule",
    onschedule_model="edc_metadata.onschedule",
    offschedule_model="edc_metadata.offschedule",
    consent_model="edc_metadata.subjectconsent",
    appointment_model="edc_appointment.appointment",
)

schedule.add_visit(visit0)
schedule.add_visit(visit1)
schedule.add_visit(visit2)
schedule.add_visit(visit3)
schedule.add_visit(visit4)

visit_schedule = VisitSchedule(
    name="visit_schedule",
    offstudy_model="edc_offstudy.subjectoffstudy",
    death_report_model="edc_metadata.deathreport",
)

visit_schedule.add_schedule(schedule)



from typing import Any, Optional

from django.apps import apps as django_apps
from edc_reference import LongitudinalRefsets, site_reference_configs


class PredicateCollection:

    """A class that groups predicates for use in rules.

    For example:

        pc = Predicates()

        @register()
        class MyRequisitionRuleGroup(RequisitionRuleGroup):

            require_cd4 = RequisitionRule(
                predicate=pc.func_require_cd4,
                consequence=REQUIRED,
                alternative=NOT_REQUIRED,
                target_panels=[cd4_panel])
    """

    app_label = "edc_metadata"
    visit_model = None

    def __init__(self) -> None:
        if not site_reference_configs.loaded:
            site_reference_configs.autodiscover()
        self.reference_model_cls = django_apps.get_model(
            site_reference_configs.get_reference_model(self.visit_model)
        )

    def values(
        self, value: Optional[Any] = None, field_name: Optional[str] = None, **kwargs
    ) -> list:
        """Returns a list of matching values or an empty list."""
        return self.exists(value=value, field_name=field_name, **kwargs)

    def exists(
        self,
        reference_name: Optional[str] = None,
        value: Optional[Any] = None,
        field_name: Optional[str] = None,
        **kwargs,
    ) -> list:
        """Returns a list of values, all or filtered, or an empty
        list.
        """
        refsets = self.refsets(reference_name=reference_name, **kwargs)
        if value:
            return refsets.fieldset(field_name).filter(value).values
        else:
            return refsets.fieldset(field_name).all().values

    def refsets(self, reference_name: Optional[str] = None, **options) -> Any:
        opts = dict(
            name=reference_name,
            visit_model=self.visit_model,
            reference_model_cls=self.reference_model_cls,
            **options,
        )
        refsets = LongitudinalRefsets(**opts)
        return refsets



# Generated by Django 2.0 on 2017-12-11 22:39

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("edc_metadata", "0007_auto_20170810_1032")]

    operations = [
        migrations.AddField(
            model_name="crfmetadata",
            name="visit_code_sequence",
            field=models.IntegerField(default=0),
        ),
        migrations.AddField(
            model_name="requisitionmetadata",
            name="visit_code_sequence",
            field=models.IntegerField(default=0),
        ),
        migrations.AlterUniqueTogether(
            name="crfmetadata",
            unique_together={
                (
                    "subject_identifier",
                    "visit_schedule_name",
                    "schedule_name",
                    "visit_code",
                    "visit_code_sequence",
                    "model",
                )
            },
        ),
        migrations.AlterUniqueTogether(
            name="requisitionmetadata",
            unique_together={
                (
                    "subject_identifier",
                    "visit_schedule_name",
                    "schedule_name",
                    "visit_code",
                    "visit_code_sequence",
                    "model",
                    "panel_name",
                )
            },
        ),
    ]



from collections import OrderedDict

from django.test import TestCase, override_settings
from edc_constants.constants import FEMALE, MALE
from edc_facility.import_holidays import import_holidays
from edc_reference.site_reference import site_reference_configs
from edc_registration.models import RegisteredSubject
from edc_utils import get_utcnow
from edc_visit_schedule.site_visit_schedules import site_visit_schedules
from edc_visit_tracking.constants import SCHEDULED
from faker import Faker

from edc_metadata.tests.reference_configs import register_to_site_reference_configs

from ...constants import NOT_REQUIRED, REQUIRED
from ...metadata_rules import (
    PF,
    CrfRule,
    CrfRuleGroup,
    CrfRuleModelConflict,
    P,
    PredicateError,
    RuleEvaluatorRegisterSubjectError,
    RuleGroupMetaError,
    site_metadata_rules,
)
from ...models import CrfMetadata
from ...target_handler import TargetModelConflict
from ..models import Appointment, CrfOne, SubjectConsent, SubjectVisit
from ..visit_schedule import visit_schedule

fake = Faker()


class CrfRuleGroupWithSourceModel(CrfRuleGroup):
    """Specifies source model."""

    crfs_male = CrfRule(
        predicate=P("f1", "eq", "car"),
        consequence=REQUIRED,
        alternative=NOT_REQUIRED,
        target_models=["crffive", "crffour"],
    )

    crfs_female = CrfRule(
        predicate=P("f1", "eq", "bicycle"),
        consequence=REQUIRED,
        alternative=NOT_REQUIRED,
        target_models=["crfthree", "crftwo"],
    )

    class Meta:
        app_label = "edc_metadata"
        source_model = "edc_metadata.crfone"


class CrfRuleGroupWithoutSourceModel(CrfRuleGroup):

    crfs_male = CrfRule(
        predicate=P("gender", "eq", MALE),
        consequence=REQUIRED,
        alternative=NOT_REQUIRED,
        target_models=["crffive", "crffour"],
    )

    crfs_female = CrfRule(
        predicate=P("gender", "eq", FEMALE),
        consequence=REQUIRED,
        alternative=NOT_REQUIRED,
        target_models=["crfthree", "crftwo"],
    )

    class Meta:
        app_label = "edc_metadata"


class CrfRuleGroupWithoutExplicitReferenceModel(CrfRuleGroup):

    crfs_male = CrfRule(
        predicate=P("gender", "eq", MALE),
        consequence=REQUIRED,
        alternative=NOT_REQUIRED,
        target_models=["crffive", "crffour"],
    )

    crfs_female = CrfRule(
        predicate=P("gender", "eq", FEMALE),
        consequence=REQUIRED,
        alternative=NOT_REQUIRED,
        target_models=["crfthree", "crftwo"],
    )

    class Meta:
        app_label = "edc_metadata"
        source_model = "edc_metadata.crfone"


class CrfRuleGroupGender(CrfRuleGroup):

    crfs_male = CrfRule(
        predicate=P("gender", "eq", MALE),
        consequence=REQUIRED,
        alternative=NOT_REQUIRED,
        target_models=["crffour", "crffive"],
    )

    crfs_female = CrfRule(
        predicate=P("gender", "eq", FEMALE),
        consequence=REQUIRED,
        alternative=NOT_REQUIRED,
        target_models=["crftwo", "crfthree"],
    )

    class Meta:
        app_label = "edc_metadata"


class TestMetadataRulesWithGender(TestCase):
    @classmethod
    def setUpTestData(cls):
        import_holidays()

    def setUp(self):
        site_reference_configs.registry = {}
        site_visit_schedules._registry = {}
        site_visit_schedules.loaded = False
        site_visit_schedules.register(visit_schedule)

        register_to_site_reference_configs()
        site_reference_configs.register_from_visit_schedule(
            visit_models={"edc_appointment.appointment": "edc_metadata.subjectvisit"}
        )

        # note crfs in visit schedule are all set to REQUIRED by default.
        _, self.schedule = site_visit_schedules.get_by_onschedule_model(
            "edc_metadata.onschedule"
        )

        site_metadata_rules.registry = OrderedDict()
        site_metadata_rules.register(rule_group_cls=CrfRuleGroupGender)

    def enroll(self, gender=None):
        subject_identifier = fake.credit_card_number()
        subject_consent = SubjectConsent.objects.create(
            subject_identifier=subject_identifier,
            consent_datetime=get_utcnow(),
            gender=gender,
        )
        self.schedule.put_on_schedule(
            subject_identifier=subject_identifier,
            onschedule_datetime=subject_consent.consent_datetime,
        )
        self.appointment = Appointment.objects.get(
            subject_identifier=subject_identifier,
            visit_code=self.schedule.visits.first.code,
        )
        subject_visit = SubjectVisit.objects.create(
            appointment=self.appointment,
            reason=SCHEDULED,
            subject_identifier=subject_identifier,
        )
        return subject_visit

    def test_rules_with_source_model(self):
        for rule in CrfRuleGroupWithSourceModel._meta.options.get("rules"):
            self.assertEqual(rule.source_model, "edc_metadata.crfone")

    def test_rules_without_source_model(self):
        for rule in CrfRuleGroupWithoutSourceModel._meta.options.get("rules"):
            self.assertIsNone(rule.source_model)

    def test_rules_source_and_reference_model_is_none(self):
        subject_visit = self.enroll(MALE)
        for rule in CrfRuleGroupWithoutSourceModel._meta.options.get("rules"):
            with self.subTest(rule=rule):
                result = rule.run(subject_visit)
                if rule.name == "crfs_male":
                    self.assertEqual(
                        result,
                        {
                            "edc_metadata.crffive": REQUIRED,
                            "edc_metadata.crffour": REQUIRED,
                        },
                    )
                elif rule.name == "crfs_female":
                    self.assertEqual(
                        result,
                        {
                            "edc_metadata.crfthree": NOT_REQUIRED,
                            "edc_metadata.crftwo": NOT_REQUIRED,
                        },
                    )

    def test_rules_with_source_but_no_explicit_reference_model(self):
        subject_visit = self.enroll(MALE)
        for rule in CrfRuleGroupWithoutExplicitReferenceModel._meta.options.get("rules"):
            with self.subTest(rule=rule):
                self.assertIsNotNone(rule.source_model)
                result = rule.run(subject_visit)
                if rule.name == "crfs_male":
                    self.assertEqual(
                        result,
                        {
                            "edc_metadata.crffive": REQUIRED,
                            "edc_metadata.crffour": REQUIRED,
                        },
                    )
                elif rule.name == "crfs_female":
                    self.assertEqual(
                        result,
                        {
                            "edc_metadata.crfthree": NOT_REQUIRED,
                            "edc_metadata.crftwo": NOT_REQUIRED,
                        },
                    )

    def test_rules_skipped_if_no_source_model_instance(self):
        subject_visit = self.enroll(MALE)
        for rule in CrfRuleGroupWithSourceModel._meta.options.get("rules"):
            with self.subTest(rule=rule):
                result = rule.run(subject_visit)
                if rule.name == "crfs_male":
                    self.assertEqual(
                        result,
                        {
                            "edc_metadata.crffive": None,
                            "edc_metadata.crffour": None,
                        },
                    )
                elif rule.name == "crfs_female":
                    self.assertEqual(
                        result,
                        {
                            "edc_metadata.crfthree": None,
                            "edc_metadata.crftwo": None,
                        },
                    )

    def test_rules_run_if_source_f1_equals_car(self):
        subject_visit = self.enroll(MALE)
        CrfOne.objects.create(subject_visit=subject_visit, f1="car")
        for rule in CrfRuleGroupWithSourceModel._meta.options.get("rules"):
            with self.subTest(rule=rule):
                result = rule.run(subject_visit)
                if rule.name == "crfs_male":
                    self.assertEqual(
                        result,
                        {
                            "edc_metadata.crffive": REQUIRED,
                            "edc_metadata.crffour": REQUIRED,
                        },
                    )
                elif rule.name == "crfs_female":
                    self.assertEqual(
                        result,
                        {
                            "edc_metadata.crfthree": NOT_REQUIRED,
                            "edc_metadata.crftwo": NOT_REQUIRED,
                        },
                    )

    def test_rules_run_if_source_f1_equals_bicycle(self):
        subject_visit = self.enroll(MALE)
        CrfOne.objects.create(subject_visit=subject_visit, f1="bicycle")
        for rule in CrfRuleGroupWithSourceModel._meta.options.get("rules"):
            with self.subTest(rule=rule):
                result = rule.run(subject_visit)
                if rule.name == "crfs_male":
                    self.assertEqual(
                        result,
                        {
                            "edc_metadata.crffive": NOT_REQUIRED,
                            "edc_metadata.crffour": NOT_REQUIRED,
                        },
                    )
                elif rule.name == "crfs_female":
                    self.assertEqual(
                        result,
                        {
                            "edc_metadata.crfthree": REQUIRED,
                            "edc_metadata.crftwo": REQUIRED,
                        },
                    )

    def test_rules_run_requires_registered_subject(self):
        subject_visit = self.enroll(MALE)
        RegisteredSubject.objects.all().delete()
        for rule in CrfRuleGroupWithSourceModel._meta.options.get("rules"):
            self.assertRaises(RuleEvaluatorRegisterSubjectError, rule.run, subject_visit)

    def test_metadata_rules_run_male_required(self):
        subject_visit = self.enroll(gender=MALE)
        for target_model in [
            "edc_metadata.crffour",
            "edc_metadata.crffive",
        ]:
            with self.subTest(target_model=target_model):
                obj = CrfMetadata.objects.get(
                    model=target_model,
                    subject_identifier=subject_visit.subject_identifier,
                    visit_code=subject_visit.visit_code,
                )
                self.assertEqual(obj.entry_status, REQUIRED)

    def test_metadata_rules_run_female_required(self):
        subject_visit = self.enroll(gender=FEMALE)
        for target_model in [
            "edc_metadata.crftwo",
            "edc_metadata.crfthree",
        ]:
            with self.subTest(target_model=target_model):
                obj = CrfMetadata.objects.get(
                    model=target_model,
                    subject_identifier=subject_visit.subject_identifier,
                    visit_code=subject_visit.visit_code,
                )
                self.assertEqual(obj.entry_status, REQUIRED)

    def test_metadata_rules_run_female_not_required(self):
        subject_visit = self.enroll(gender=FEMALE)
        for target_model in [
            "edc_metadata.crffour",
            "edc_metadata.crffive",
        ]:
            with self.subTest(target_model=target_model):
                obj = CrfMetadata.objects.get(
                    model=target_model,
                    subject_identifier=subject_visit.subject_identifier,
                    visit_code=subject_visit.visit_code,
                )
                self.assertEqual(obj.entry_status, NOT_REQUIRED)

    def test_metadata_rules_run_male_not_required(self):
        subject_visit = self.enroll(gender=MALE)
        for target_model in [
            "edc_metadata.crftwo",
            "edc_metadata.crfthree",
        ]:
            with self.subTest(target_model=target_model):
                obj = CrfMetadata.objects.get(
                    model=target_model,
                    subject_identifier=subject_visit.subject_identifier,
                    visit_code=subject_visit.visit_code,
                )
                self.assertEqual(obj.entry_status, NOT_REQUIRED)

    def test_rule_group_metadata_objects(self):
        subject_visit = self.enroll(gender=MALE)
        _, metadata_objects = CrfRuleGroupGender().evaluate_rules(visit=subject_visit)
        self.assertEqual(metadata_objects.get("edc_metadata.crffour").entry_status, REQUIRED)
        self.assertEqual(metadata_objects.get("edc_metadata.crffive").entry_status, REQUIRED)
        self.assertEqual(
            metadata_objects.get("edc_metadata.crftwo").entry_status, NOT_REQUIRED
        )
        self.assertEqual(
            metadata_objects.get("edc_metadata.crfthree").entry_status,
            NOT_REQUIRED,
        )

    def test_rule_group_rule_results(self):
        subject_visit = self.enroll(gender=MALE)
        rule_results, _ = CrfRuleGroupGender().evaluate_rules(visit=subject_visit)
        self.assertEqual(
            rule_results["CrfRuleGroupGender.crfs_male"].get("edc_metadata.crffour"),
            REQUIRED,
        )
        self.assertEqual(
            rule_results["CrfRuleGroupGender.crfs_male"].get("edc_metadata.crffive"),
            REQUIRED,
        )
        self.assertEqual(
            rule_results["CrfRuleGroupGender.crfs_female"].get("edc_metadata.crftwo"),
            NOT_REQUIRED,
        )
        self.assertEqual(
            rule_results["CrfRuleGroupGender.crfs_female"].get("edc_metadata.crfthree"),
            NOT_REQUIRED,
        )

    def test_bad_rule_group_target_model_cannot_also_be_source_model(self):

        site_metadata_rules.registry = OrderedDict()
        subject_visit = self.enroll(gender=MALE)

        class BadCrfRuleGroup(CrfRuleGroup):
            crfs_male = CrfRule(
                predicate=P("f1", "eq", "car"),
                consequence=REQUIRED,
                alternative=NOT_REQUIRED,
                target_models=["crfone"],
            )

            class Meta:
                app_label = "edc_metadata"
                source_model = "edc_metadata.crfone"

        self.assertRaises(
            CrfRuleModelConflict, BadCrfRuleGroup().evaluate_rules, visit=subject_visit
        )

    def test_rule_group_target_model_cannot_be_visit_model(self):
        site_metadata_rules.registry = OrderedDict()

        subject_visit = self.enroll(gender=MALE)

        class BadCrfRuleGroup(CrfRuleGroup):
            rule = CrfRule(
                predicate=P("f1", "eq", "car"),
                consequence=REQUIRED,
                alternative=NOT_REQUIRED,
                target_models=["subjectvisit"],
            )

            class Meta:
                app_label = "edc_metadata"
                source_model = "edc_metadata.crfone"

        self.assertRaises(
            TargetModelConflict, BadCrfRuleGroup().evaluate_rules, visit=subject_visit
        )

    def test_bad_predicate_blah_is_not_an_operator(self):
        try:

            class BadCrfRuleGroup(CrfRuleGroup):
                rule = CrfRule(
                    predicate=P("f1", "blah", "car"),
                    consequence=REQUIRED,
                    alternative=NOT_REQUIRED,
                    target_models=["crftwo"],
                )

                class Meta:
                    app_label = "edc_metadata"
                    source_model = "edc_metadata.crfone"

        except PredicateError:
            pass
        else:
            self.fail("PredicateError not raised.")

    def test_predicate_pf(self):
        """Asserts entry status unchanged if source model does not exist."""

        def func(f1, f2):
            return True if f1 == "f1" and f2 == "f2" else False

        class MyCrfRuleGroup(CrfRuleGroup):
            rule = CrfRule(
                predicate=PF("f1", "f2", func=func),
                consequence=REQUIRED,
                alternative=NOT_REQUIRED,
                target_models=["crftwo"],
            )

            class Meta:
                app_label = "edc_metadata"
                source_model = "edc_metadata.crfone"

        site_metadata_rules.registry = OrderedDict()
        subject_visit = self.enroll(gender=MALE)

        MyCrfRuleGroup().evaluate_rules(visit=subject_visit)

        obj = CrfMetadata.objects.get(
            model="edc_metadata.crftwo",
            subject_identifier=subject_visit.subject_identifier,
            visit_code=subject_visit.visit_code,
        )
        self.assertEqual(obj.entry_status, REQUIRED)

    @override_settings(DEBUG=True)
    def test_predicate_pf2(self):
        """Asserts entry status set to alternative if source model
        exists but does not meet criteria.
        """

        def func(f1, f2):
            return f1 == "f1_value" and f2 == "f2_value"

        class MyCrfRuleGroup(CrfRuleGroup):
            rule = CrfRule(
                predicate=PF("f1", "f2", func=func),
                consequence=REQUIRED,
                alternative=NOT_REQUIRED,
                target_models=["crftwo"],
            )

            class Meta:
                app_label = "edc_metadata"
                source_model = "edc_metadata.crfone"

        site_metadata_rules.registry = OrderedDict()
        site_metadata_rules.register(MyCrfRuleGroup)
        subject_visit = self.enroll(gender=MALE)
        obj = CrfMetadata.objects.get(
            model="edc_metadata.crftwo",
            subject_identifier=subject_visit.subject_identifier,
            visit_code=subject_visit.visit_code,
        )
        self.assertEqual(obj.entry_status, REQUIRED)
        CrfOne.objects.create(subject_visit=subject_visit, f1="blah", f2="f2_value")
        self.assertFalse(func("blah", "f2_value"))
        # MyCrfRuleGroup().evaluate_rules(visit=subject_visit)
        obj = CrfMetadata.objects.get(
            model="edc_metadata.crftwo",
            subject_identifier=subject_visit.subject_identifier,
            visit_code=subject_visit.visit_code,
        )

        self.assertEqual(obj.entry_status, NOT_REQUIRED)

    def test_predicate_pf3(self):
        """Asserts entry status set to consequence if source model
        exists and meets criteria.
        """

        def func(f1, f2):
            return True if f1 == "f1" and f2 == "f2" else False

        class MyCrfRuleGroup(CrfRuleGroup):
            rule = CrfRule(
                predicate=PF("f1", "f2", func=func),
                consequence=REQUIRED,
                alternative=NOT_REQUIRED,
                target_models=["crftwo"],
            )

            class Meta:
                app_label = "edc_metadata"
                source_model = "edc_metadata.crfone"

        site_metadata_rules.registry = OrderedDict()
        subject_visit = self.enroll(gender=MALE)
        CrfOne.objects.create(subject_visit=subject_visit, f1="f1", f2="f2")
        MyCrfRuleGroup().evaluate_rules(visit=subject_visit)

        obj = CrfMetadata.objects.get(
            model="edc_metadata.crftwo",
            subject_identifier=subject_visit.subject_identifier,
            visit_code=subject_visit.visit_code,
        )
        self.assertEqual(obj.entry_status, REQUIRED)

    def test_p_repr(self):

        p = P("blah", "eq", "car")
        self.assertTrue(repr(p))

    def test_pf_repr(self):
        def func(f1, f2):
            return True if f1 == "f1" and f2 == "f2" else False

        pf = PF("blah", "f2", func=func)
        self.assertTrue(repr(pf))

    def test_rule_repr(self):
        rule = CrfRule(
            predicate=P("f1", "eq", "car"),
            consequence=REQUIRED,
            alternative=NOT_REQUIRED,
            target_models=["crftwo"],
        )
        self.assertTrue(repr(rule))

    def test_rule_group_meta_repr(self):
        class MyCrfRuleGroup(CrfRuleGroup):
            rule = CrfRule(
                predicate=P("f1", "eq", "car"),
                consequence=REQUIRED,
                alternative=NOT_REQUIRED,
                target_models=["crftwo"],
            )

            class Meta:
                app_label = "edc_metadata"
                source_model = "edc_metadata.crfmissingmanager"

        self.assertTrue(repr(MyCrfRuleGroup()._meta))

    def test_sub_class_rule_group(self):
        class MyCrfRuleGroup(CrfRuleGroup):
            rule1 = CrfRule(
                predicate=P("f1", "eq", "car"),
                consequence=REQUIRED,
                alternative=NOT_REQUIRED,
                target_models=["crftwo"],
            )

            class Meta:
                abstract = True

        class NewCrfRuleGroup(CrfRuleGroup):

            rule2 = CrfRule(
                predicate=P("f1", "eq", "car"),
                consequence=REQUIRED,
                alternative=NOT_REQUIRED,
                target_models=["crfthree"],
            )

            class Meta:
                app_label = "edc_metadata"
                source_model = "edc_metadata.crfmissingmanager"

        self.assertTrue(len(NewCrfRuleGroup()._meta.options.get("rules")), 2)

    def test_rule_group_missing_meta(self):

        try:

            class MyCrfRuleGroup(CrfRuleGroup):
                rule1 = CrfRule(
                    predicate=P("f1", "eq", "car"),
                    consequence=REQUIRED,
                    alternative=NOT_REQUIRED,
                    target_models=["crftwo"],
                )

        except AttributeError:
            pass
        else:
            self.fail("AttributeError not raised.")

    def test_rule_group_invalid_meta_option(self):

        try:

            class MyCrfRuleGroup(CrfRuleGroup):
                rule1 = CrfRule(
                    predicate=P("f1", "eq", "car"),
                    consequence=REQUIRED,
                    alternative=NOT_REQUIRED,
                    target_models=["crftwo"],
                )

                class Meta:
                    app_label = "edc_metadata"
                    source_model = "edc_metadata.crfmissingmanager"
                    blah = "blah"

        except RuleGroupMetaError:
            pass
        else:
            self.fail("AttributeError not raised.")



# Generated by Django 3.0.4 on 2020-05-12 21:23

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("edc_metadata", "0017_auto_20191024_1000"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="crfmetadata",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "get_latest_by": "modified",
                "ordering": ("show_order",),
                "verbose_name": "Crf Metadata",
                "verbose_name_plural": "Crf Metadata",
            },
        ),
        migrations.AlterModelOptions(
            name="requisitionmetadata",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "get_latest_by": "modified",
                "ordering": ("show_order",),
                "verbose_name": "Requisition Metadata",
                "verbose_name_plural": "Requisition Metadata",
            },
        ),
    ]



from __future__ import annotations

from ..metadata import RequisitionMetadataGetter
from .metadata_wrappers import MetadataWrappers
from .requisition_metadata_wrapper import RequisitionMetadataWrapper


class RequisitionMetadataWrappers(MetadataWrappers):

    metadata_getter_cls: RequisitionMetadataGetter = RequisitionMetadataGetter
    metadata_wrapper_cls: RequisitionMetadataWrapper = RequisitionMetadataWrapper



from django.db import models


class CrfMetadataManager(models.Manager):

    use_in_migrations = True

    def get_by_natural_key(
        self,
        model,
        subject_identifier,
        schedule_name,
        visit_schedule_name,
        visit_code,
        visit_code_sequence,
    ):

        return self.get(
            model=model,
            subject_identifier=subject_identifier,
            schedule_name=schedule_name,
            visit_schedule_name=visit_schedule_name,
            visit_code=visit_code,
            visit_code_sequence=visit_code_sequence,
        )


class RequisitionMetadataManager(models.Manager):

    use_in_migrations = True

    def get_by_natural_key(
        self,
        panel_name,
        model,
        subject_identifier,
        visit_schedule_name,
        schedule_name,
        visit_code,
        visit_code_sequence,
    ):
        return self.get(
            panel_name=panel_name,
            model=model,
            subject_identifier=subject_identifier,
            schedule_name=schedule_name,
            visit_schedule_name=visit_schedule_name,
            visit_code=visit_code,
            visit_code_sequence=visit_code_sequence,
        )



from __future__ import annotations

from typing import Type

from django.db import models

from .constants import KEYED
from .target_handler import TargetHandler


class MetadataUpdaterError(Exception):
    pass


class MetadataUpdater:
    """A class to update a subject's metadata given
    the related_visit, target model name and desired entry status.
    """

    target_handler = TargetHandler

    def __init__(self, related_visit=None, target_model=None):
        self._metadata_obj = None
        self.related_visit = related_visit
        self.target_model = target_model

    def __repr__(self):
        return (
            f"{self.__class__.__name__}"
            f"(related_visit={self.related_visit}, "
            f"target_model={self.target_model})"
        )

    def update(self, entry_status: str = None) -> Type[models.Model]:
        metadata_obj = self.target.metadata_obj
        if self.target.object:
            entry_status = KEYED
        if entry_status and metadata_obj.entry_status != entry_status:
            metadata_obj.entry_status = entry_status
            metadata_obj.save()
            metadata_obj.refresh_from_db()
            if metadata_obj.entry_status != entry_status:
                raise MetadataUpdaterError(
                    "Expected entry status does not match `entry_status` on "
                    "metadata model instance. "
                    f"Got {entry_status} != {metadata_obj.entry_status}."
                )
        return metadata_obj

    @property
    def target(self):
        return self.target_handler(model=self.target_model, related_visit=self.related_visit)



from .crf_metadata_wrapper import CrfMetadataWrapper
from .crf_metadata_wrappers import CrfMetadataWrappers
from .metadata_wrapper import MetadataWrapper, MetadataWrapperError
from .requisition_metadata_wrapper import RequisitionMetadataWrapper
from .requisition_metadata_wrappers import RequisitionMetadataWrappers



# -*- coding: utf-8 -*-
# Generated by Django 1.10.4 on 2017-02-09 07:24
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("edc_metadata", "0005_auto_20170112_0602")]

    operations = [
        migrations.AlterField(
            model_name="crfmetadata",
            name="schedule_name",
            field=models.CharField(editable=False, max_length=25),
        ),
        migrations.AlterField(
            model_name="crfmetadata",
            name="visit_schedule_name",
            field=models.CharField(
                editable=False,
                help_text='the name of the visit schedule used to find the "schedule"',
                max_length=25,
            ),
        ),
        migrations.AlterField(
            model_name="requisitionmetadata",
            name="schedule_name",
            field=models.CharField(editable=False, max_length=25),
        ),
        migrations.AlterField(
            model_name="requisitionmetadata",
            name="visit_schedule_name",
            field=models.CharField(
                editable=False,
                help_text='the name of the visit schedule used to find the "schedule"',
                max_length=25,
            ),
        ),
    ]



import sys
from typing import Any, Optional

from django.apps import apps as django_apps
from django.core.management.color import color_style
from edc_reference.site_reference import site_reference_configs

style = color_style()


class RuleGroup:

    """Base class for CRF and Requisition rule groups."""

    @classmethod
    def get_rules(cls: Any) -> Any:
        return cls._meta.options.get("rules")

    @classmethod
    def validate(cls: Any) -> None:
        """Outputs to the console if a target model referenced in a rule
        does not exist.
        """
        # TODO: extend this list
        default_fields = ["gender"]

        # verify models exists
        if cls._meta.source_model:
            cls._lookup_model(model=cls._meta.source_model, category="source")

        for rule in cls.get_rules():
            for target_model in rule.target_models:
                cls._lookup_model(model=target_model, category="target")

        # verify fields referred to on source models
        # note: fields referenced in funcs in a predicate collection
        # are not verified here.
        if cls._meta.source_model:
            model_cls = cls._lookup_model(model=cls._meta.source_model, category="source")
            fields = [f.name for f in model_cls._meta.get_fields()]
            fields.extend(default_fields)
            for rule in cls.get_rules():
                for field_name in rule.field_names:
                    if field_name not in fields:
                        sys.stdout.write(
                            style.ERROR(
                                f"  (?) Field {cls._meta.source_model}.{field_name} "
                                f"is invalid.\n"
                            )
                        )
                    reference_fields = site_reference_configs.get_fields(
                        name=cls._meta.source_model
                    )
                    reference_fields.extend(default_fields)
                    if field_name not in reference_fields:
                        sys.stdout.write(
                            style.ERROR(
                                f"  (?) Field {cls._meta.source_model}.{field_name} "
                                f"not found in site_reference_configs.\n"
                            )
                        )

    @classmethod
    def _lookup_model(cls, model: Optional[str] = None, category: Optional[str] = None) -> Any:
        sys.stdout.write(f"  ( ) {model}\r")
        model_cls = None
        try:
            model_cls = django_apps.get_model(model)
        except LookupError:
            sys.stdout.write(style.ERROR(f"  (?) {model}. See {category} model in {cls}\n"))
        else:
            sys.stdout.write(f"  (*) {model}\n")
        return model_cls



from datetime import date

from django.db import models
from django.db.models.deletion import PROTECT
from edc_appointment.models import Appointment
from edc_consent.field_mixins import PersonalFieldsMixin
from edc_constants.choices import YES_NO
from edc_constants.constants import MALE
from edc_crf.model_mixins import CrfWithActionModelMixin
from edc_identifier.managers import SubjectIdentifierManager
from edc_identifier.model_mixins import (
    UniqueSubjectIdentifierFieldMixin,
    UniqueSubjectIdentifierModelMixin,
)
from edc_lab.model_mixins import PanelModelMixin
from edc_list_data.model_mixins import ListModelMixin
from edc_model.models import BaseUuidModel
from edc_reference.model_mixins import (
    ReferenceModelMixin,
    RequisitionReferenceModelMixin,
)
from edc_registration.model_mixins import UpdatesOrCreatesRegistrationModelMixin
from edc_sites.models import SiteModelMixin
from edc_utils import get_utcnow
from edc_visit_schedule.model_mixins import OffScheduleModelMixin, OnScheduleModelMixin
from edc_visit_tracking.model_mixins import (
    SubjectVisitMissedModelMixin,
    VisitModelMixin,
    VisitTrackingCrfModelMixin,
)

from ..model_mixins.creates import CreatesMetadataModelMixin
from ..model_mixins.updates import (
    UpdatesCrfMetadataModelMixin,
    UpdatesRequisitionMetadataModelMixin,
)


class CrfModelMixin(VisitTrackingCrfModelMixin, SiteModelMixin, models.Model):
    class Meta:
        abstract = True


class OnSchedule(OnScheduleModelMixin, BaseUuidModel):

    pass


class OffSchedule(OffScheduleModelMixin, BaseUuidModel):

    pass


class DeathReport(UniqueSubjectIdentifierFieldMixin, BaseUuidModel):

    objects = SubjectIdentifierManager()

    def natural_key(self):
        return (self.subject_identifier,)  # noqa


class SubjectConsent(
    UniqueSubjectIdentifierModelMixin,
    PersonalFieldsMixin,
    UpdatesOrCreatesRegistrationModelMixin,
    SiteModelMixin,
    BaseUuidModel,
):

    consent_datetime = models.DateTimeField(default=get_utcnow)

    version = models.CharField(max_length=25, default="1")

    screening_identifier = models.CharField(max_length=25, null=True)

    identity = models.CharField(max_length=25, default="111111111")

    confirm_identity = models.CharField(max_length=25, default="111111111")

    dob = models.DateField(default=date(1995, 1, 1))

    gender = models.CharField(max_length=25, default=MALE)

    objects = SubjectIdentifierManager()

    def natural_key(self):
        return (self.subject_identifier,)  # noqa

    class Meta(BaseUuidModel.Meta):
        verbose_name = "Subject Consent"
        verbose_name_plural = "Subject Consents"


class SubjectVisit(
    VisitModelMixin,
    ReferenceModelMixin,
    CreatesMetadataModelMixin,
    SiteModelMixin,
    BaseUuidModel,
):

    appointment = models.OneToOneField(Appointment, on_delete=PROTECT)

    subject_identifier = models.CharField(max_length=50)

    reason = models.CharField(max_length=25)


class SubjectRequisition(
    CrfModelMixin,
    RequisitionReferenceModelMixin,
    PanelModelMixin,
    UpdatesRequisitionMetadataModelMixin,
    BaseUuidModel,
):

    subject_visit = models.ForeignKey(SubjectVisit, on_delete=PROTECT)

    requisition_datetime = models.DateTimeField(null=True)

    is_drawn = models.CharField(max_length=25, choices=YES_NO, null=True)

    reason_not_drawn = models.CharField(max_length=25, null=True)


class CrfOne(CrfModelMixin, ReferenceModelMixin, UpdatesCrfMetadataModelMixin, BaseUuidModel):

    subject_visit = models.ForeignKey(SubjectVisit, on_delete=PROTECT)

    f1 = models.CharField(max_length=50, null=True)

    f2 = models.CharField(max_length=50, null=True)

    f3 = models.CharField(max_length=50, null=True)


class CrfTwo(CrfModelMixin, ReferenceModelMixin, UpdatesCrfMetadataModelMixin, BaseUuidModel):

    subject_visit = models.ForeignKey(SubjectVisit, on_delete=PROTECT)

    f1 = models.CharField(max_length=50, null=True)


class CrfThree(
    CrfModelMixin, ReferenceModelMixin, UpdatesCrfMetadataModelMixin, BaseUuidModel
):

    subject_visit = models.ForeignKey(SubjectVisit, on_delete=PROTECT)

    f1 = models.CharField(max_length=50, null=True)


class CrfFour(CrfModelMixin, ReferenceModelMixin, UpdatesCrfMetadataModelMixin, BaseUuidModel):

    subject_visit = models.ForeignKey(SubjectVisit, on_delete=PROTECT)

    f1 = models.CharField(max_length=50, null=True)


class CrfFive(CrfModelMixin, ReferenceModelMixin, UpdatesCrfMetadataModelMixin, BaseUuidModel):

    subject_visit = models.ForeignKey(SubjectVisit, on_delete=PROTECT)

    f1 = models.CharField(max_length=50, null=True)


class Crfsix(CrfModelMixin, ReferenceModelMixin, UpdatesCrfMetadataModelMixin, BaseUuidModel):

    subject_visit = models.ForeignKey(SubjectVisit, on_delete=PROTECT)

    f1 = models.CharField(max_length=50, null=True)


class CrfSeven(
    CrfModelMixin, ReferenceModelMixin, UpdatesCrfMetadataModelMixin, BaseUuidModel
):

    subject_visit = models.ForeignKey(SubjectVisit, on_delete=PROTECT)

    f1 = models.CharField(max_length=50, null=True)


class PrnOne(CrfModelMixin, ReferenceModelMixin, UpdatesCrfMetadataModelMixin, BaseUuidModel):

    subject_visit = models.ForeignKey(SubjectVisit, on_delete=PROTECT)

    f1 = models.CharField(max_length=50, null=True)


class PrnTwo(CrfModelMixin, ReferenceModelMixin, UpdatesCrfMetadataModelMixin, BaseUuidModel):

    subject_visit = models.ForeignKey(SubjectVisit, on_delete=PROTECT)

    f1 = models.CharField(max_length=50, null=True)


class MissedVisit(
    CrfModelMixin, ReferenceModelMixin, UpdatesCrfMetadataModelMixin, BaseUuidModel
):

    subject_visit = models.ForeignKey(SubjectVisit, on_delete=PROTECT)

    f1 = models.CharField(max_length=50, null=True)


class CrfMissingManager(ReferenceModelMixin, BaseUuidModel):

    subject_visit = models.ForeignKey(SubjectVisit, on_delete=PROTECT)

    f1 = models.CharField(max_length=50, null=True)


class SubjectVisitMissedReasons(ListModelMixin):
    class Meta(ListModelMixin.Meta):
        verbose_name = "Subject Missed Visit Reasons"
        verbose_name_plural = "Subject Missed Visit Reasons"


class SubjectVisitMissed(
    SubjectVisitMissedModelMixin,
    CrfWithActionModelMixin,
    BaseUuidModel,
):

    missed_reasons = models.ManyToManyField(
        SubjectVisitMissedReasons, blank=True, related_name="+"
    )

    class Meta(
        SubjectVisitMissedModelMixin.Meta,
        BaseUuidModel.Meta,
    ):
        verbose_name = "Missed Visit Report"
        verbose_name_plural = "Missed Visit Report"



from __future__ import annotations

from typing import TYPE_CHECKING

from django.apps import apps as django_apps
from edc_reference import site_reference_configs
from edc_visit_schedule.site_visit_schedules import site_visit_schedules

from ..constants import REQUISITION
from ..target_handler import TargetHandler
from .requisition_metadata_handler import RequisitionMetadataHandler

if TYPE_CHECKING:
    from edc_lab.models import Panel


class TargetPanelNotScheduledForVisit(Exception):
    pass


class InvalidTargetPanel(Exception):
    pass


class RequisitionTargetHandler(TargetHandler):
    metadata_handler_cls: RequisitionMetadataHandler = RequisitionMetadataHandler
    metadata_category: str = REQUISITION
    metadata_model: str = "edc_metadata.requisitionmetadata"

    def __init__(self, target_panel: Panel = None, **kwargs):
        self.target_panel = target_panel
        super().__init__(**kwargs)

    @property
    def reference_model_cls(self):
        name = f"{self.model}.{self.target_panel.name}"
        reference_model = site_reference_configs.get_reference_model(name=name)
        return django_apps.get_model(reference_model)

    @property
    def object(self):
        return self.reference_model_cls.objects.get_requisition_for_visit(
            visit=self.related_visit,
            name=f"{self.model}.{self.target_panel.name}",
        )

    @property
    def target_panel_names(self):
        """Returns a list of panels for this visit."""
        if self.related_visit.visit_code_sequence != 0:
            forms = (
                self.related_visit.visit.requisitions_unscheduled
                + self.related_visit.visit.requisitions_prn
            )
        else:
            forms = (
                self.related_visit.visit.requisitions
                + self.related_visit.visit.requisitions_prn
            )
        return list(set([form.panel.name for form in forms]))

    def raise_on_not_scheduled_for_visit(self):
        """Raises an exception if target_panel is not scheduled
        for this visit or is invalid.
        """
        self.raise_on_invalid_panel()
        if self.target_panel.name not in self.target_panel_names:
            raise TargetPanelNotScheduledForVisit(
                f"Target panel {self.target_panel.name} is not scheduled "
                f"for visit '{self.related_visit.visit_code}'."
            )

    @property
    def schedule(self):
        """Returns a schedule instance from site_visit_schedule
        for this visit.
        """
        visit_schedule = site_visit_schedules.get_visit_schedule(
            self.related_visit.visit_schedule_name
        )
        return visit_schedule.schedules.get(self.related_visit.schedule_name)

    def raise_on_invalid_panel(self):
        """Raises an exception if target_panel is not found in any visit
        for this schedule.
        """
        panel_names = []
        for visit in self.schedule.visits.values():
            panel_names.extend([r.panel.name for r in visit.all_requisitions])
        if self.target_panel.name not in panel_names:
            raise InvalidTargetPanel(
                f"Invalid panel. {self.target_panel.name} is not a valid "
                f"panel for any visit in schedule {repr(self.schedule)}. "
            )

    @property
    def metadata_handler(self):
        return self.metadata_handler_cls(
            metadata_model=self.metadata_model,
            model=self.model,
            related_visit=self.related_visit,
            panel=self.target_panel,
        )



class RefreshMetadataViewMixin:
    """
    Declare together with the edc_appointment.AppointmentViewMixin.

    """

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        return context



from __future__ import annotations

from typing import TYPE_CHECKING, Any
from warnings import warn

from django.apps import apps as django_apps
from django.contrib.admin.sites import all_sites
from django.core.exceptions import (
    ImproperlyConfigured,
    MultipleObjectsReturned,
    ObjectDoesNotExist,
)

from ..constants import KEYED
from .metadata import model_cls_registered_with_admin_site

if TYPE_CHECKING:
    from edc_appointment.models import Appointment
    from edc_visit_tracking.model_mixins import VisitModelMixin

    from ..models import CrfMetadata, RequisitionMetadata


class MetadataGetterError(Exception):
    pass


class MetadataValidator:
    def __init__(self, metadata_obj: Any, related_visit: Any) -> None:
        self.metadata_obj = metadata_obj
        self.related_visit = related_visit
        self.validate_metadata_object()

    @property
    def extra_query_attrs(self) -> dict:
        return {}

    def validate_metadata_object(self) -> None:
        if self.metadata_obj:
            # confirm model class exists
            try:
                model_cls = django_apps.get_model(self.metadata_obj.model)
            except LookupError:
                self.metadata_obj.delete()
                self.metadata_obj = None
            else:
                if not model_cls_registered_with_admin_site(model_cls):
                    warn(
                        "Model class not registered with Admin. "
                        f"Deleting related metadata. Got {model_cls}."
                    )
                    # TODO: delete all metadata instances for this model cls?
                    self.metadata_obj.delete()
                    self.metadata_obj = None
                else:
                    # confirm metadata.entry_status is correct
                    model_obj = None
                    query_attrs = {
                        f"{model_cls.related_visit_model_attr()}": self.related_visit
                    }
                    query_attrs.update(**self.extra_query_attrs)
                    try:
                        model_obj = model_cls.objects.get(**query_attrs)
                    except AttributeError as e:
                        if "related_visit_model_attr" not in str(e):
                            raise ImproperlyConfigured(f"{e} See {repr(model_cls)}")
                        raise
                    except ObjectDoesNotExist:
                        pass
                    except MultipleObjectsReturned:
                        raise
                    self.verify_entry_status_with_model_obj(model_obj)

    def verify_entry_status_with_model_obj(self, model_obj: Any) -> None:
        """Verifies that entry_status is set to KEYED if the model
        instance exists, etc.

        Fixes on the fly if model obj exists and entry status != KEYED."""
        for i in range(2):
            # TODO: add test, fixes on the fly
            if model_obj and model_obj.id and self.metadata_obj.entry_status != KEYED:
                self.metadata_obj.entry_status = KEYED
                self.metadata_obj.save()
                self.metadata_obj.refresh_from_db()
                break
            # TODO: add test, how can the entry status be reset?
            #  Do we need to recalculate the visit's metadata??
            # what about metadata rules?
            elif not model_obj and self.metadata_obj.entry_status == KEYED:
                if self.related_visit:
                    # resave subject visit to recreate / calculate metadata
                    self.related_visit.save()
                    self.metadata_obj.refresh_from_db()
                    continue

    @staticmethod
    def model_cls_registered_with_admin_site(model_cls: Any) -> bool:
        """Returns True if model cls is registered in Admin."""
        for admin_site in all_sites:
            if model_cls in admin_site._registry:
                return True
        return False


class MetadataGetter:

    """A class that gets a filtered queryset of metadata --
    `metadata_objects`.
    """

    metadata_model: str = None

    metadata_validator_cls = MetadataValidator

    def __init__(self, appointment: Appointment) -> None:
        self.options = {}
        self.appointment = appointment
        self.related_visit: VisitModelMixin | None = getattr(
            self.appointment, "related_visit", None
        )
        instance = self.related_visit or self.appointment
        self.subject_identifier = instance.subject_identifier
        self.visit_code = instance.visit_code
        self.visit_code_sequence = instance.visit_code_sequence
        query_options = dict(
            subject_identifier=self.subject_identifier,
            visit_code=self.visit_code,
            visit_code_sequence=self.visit_code_sequence,
            visit_schedule_name=instance.visit_schedule_name,
            schedule_name=instance.schedule_name,
        )
        self.validate_metadata_objects(query_options)
        self.metadata_objects = self.metadata_model_cls.objects.filter(
            **query_options
        ).order_by("show_order")

    @property
    def metadata_model_cls(self) -> CrfMetadata | RequisitionMetadata:
        return django_apps.get_model(self.metadata_model)

    def next_object(
        self, show_order: int | None = None, entry_status: str | None = None
    ) -> CrfMetadata | RequisitionMetadata:
        """Returns the next model instance based on the show order."""
        if show_order is None:
            metadata_obj = None
        else:
            opts = {"show_order__gt": show_order}
            if entry_status:
                opts.update(entry_status=entry_status)
            metadata_obj = self.metadata_objects.filter(**opts).order_by("show_order").first()
        return metadata_obj

    def validate_metadata_objects(self, query_options: dict) -> None:
        qs = self.metadata_model_cls.objects.filter(**query_options)
        for metadata_obj in qs:
            self.metadata_validator_cls(metadata_obj, self.related_visit)



from django.core.exceptions import ObjectDoesNotExist
from django.test import TestCase
from edc_visit_tracking.constants import SCHEDULED

from ...constants import KEYED, NOT_REQUIRED, REQUIRED
from ...metadata_inspector import MetaDataInspector
from ...metadata_updater import MetadataUpdater
from ...models import CrfMetadata, RequisitionMetadata
from ...target_handler import TargetModelLookupError, TargetModelNotScheduledForVisit
from ..models import CrfOne, CrfThree, CrfTwo, SubjectRequisition, SubjectVisit
from .metadata_test_mixin import TestMetadataMixin


class TestMetadataUpdater(TestMetadataMixin, TestCase):
    def test_updates_crf_metadata_as_keyed(self):
        subject_visit = SubjectVisit.objects.create(
            appointment=self.appointment, reason=SCHEDULED
        )
        CrfOne.objects.create(subject_visit=subject_visit)
        self.assertEqual(
            CrfMetadata.objects.filter(
                entry_status=KEYED,
                model="edc_metadata.crfone",
                visit_code=subject_visit.visit_code,
            ).count(),
            1,
        )
        self.assertEqual(
            CrfMetadata.objects.filter(
                entry_status=REQUIRED,
                model="edc_metadata.crftwo",
                visit_code=subject_visit.visit_code,
            ).count(),
            1,
        )
        self.assertEqual(
            CrfMetadata.objects.filter(
                entry_status=REQUIRED,
                model="edc_metadata.crfthree",
                visit_code=subject_visit.visit_code,
            ).count(),
            1,
        )

    def test_updates_all_crf_metadata_as_keyed(self):
        subject_visit = SubjectVisit.objects.create(
            appointment=self.appointment, reason=SCHEDULED
        )
        CrfOne.objects.create(subject_visit=subject_visit)
        CrfTwo.objects.create(subject_visit=subject_visit)
        CrfThree.objects.create(subject_visit=subject_visit)
        for model_name in ["crfone", "crftwo", "crfthree"]:
            self.assertEqual(
                CrfMetadata.objects.filter(
                    entry_status=KEYED,
                    model=f"edc_metadata.{model_name}",
                    visit_code=subject_visit.visit_code,
                ).count(),
                1,
            )

    def test_updates_requisition_metadata_as_keyed(self):
        subject_visit = SubjectVisit.objects.create(
            appointment=self.appointment, reason=SCHEDULED
        )
        SubjectRequisition.objects.create(subject_visit=subject_visit, panel=self.panel_one)
        self.assertEqual(
            RequisitionMetadata.objects.filter(
                entry_status=KEYED,
                model="edc_metadata.subjectrequisition",
                panel_name=self.panel_one.name,
                visit_code=subject_visit.visit_code,
            ).count(),
            1,
        )
        self.assertEqual(
            RequisitionMetadata.objects.filter(
                entry_status=REQUIRED,
                model="edc_metadata.subjectrequisition",
                panel_name=self.panel_two.name,
                visit_code=subject_visit.visit_code,
            ).count(),
            1,
        )

    def test_resets_crf_metadata_on_delete(self):
        subject_visit = SubjectVisit.objects.create(
            appointment=self.appointment, reason=SCHEDULED
        )
        crf_one = CrfOne.objects.create(subject_visit=subject_visit)
        crf_one.delete()
        self.assertEqual(
            CrfMetadata.objects.filter(
                entry_status=REQUIRED,
                model="edc_metadata.crfone",
                visit_code=subject_visit.visit_code,
            ).count(),
            1,
        )
        self.assertEqual(
            CrfMetadata.objects.filter(
                entry_status=REQUIRED,
                model="edc_metadata.crftwo",
                visit_code=subject_visit.visit_code,
            ).count(),
            1,
        )
        self.assertEqual(
            CrfMetadata.objects.filter(
                entry_status=REQUIRED,
                model="edc_metadata.crfthree",
                visit_code=subject_visit.visit_code,
            ).count(),
            1,
        )

    def test_resets_requisition_metadata_on_delete1(self):
        subject_visit = SubjectVisit.objects.create(
            appointment=self.appointment, reason=SCHEDULED
        )
        obj = SubjectRequisition.objects.create(
            subject_visit=subject_visit, panel=self.panel_one
        )
        obj.delete()
        self.assertEqual(
            RequisitionMetadata.objects.filter(
                entry_status=REQUIRED,
                model="edc_metadata.subjectrequisition",
                panel_name=self.panel_one.name,
                visit_code=subject_visit.visit_code,
            ).count(),
            1,
        )
        self.assertEqual(
            RequisitionMetadata.objects.filter(
                entry_status=REQUIRED,
                model="edc_metadata.subjectrequisition",
                panel_name=self.panel_two.name,
                visit_code=subject_visit.visit_code,
            ).count(),
            1,
        )

    def test_resets_requisition_metadata_on_delete2(self):
        subject_visit = SubjectVisit.objects.create(
            appointment=self.appointment, reason=SCHEDULED
        )
        obj = SubjectRequisition.objects.create(
            subject_visit=subject_visit, panel=self.panel_two
        )
        obj.delete()
        self.assertEqual(
            RequisitionMetadata.objects.filter(
                entry_status=REQUIRED,
                model="edc_metadata.subjectrequisition",
                panel_name=self.panel_one.name,
                visit_code=subject_visit.visit_code,
            ).count(),
            1,
        )
        self.assertEqual(
            RequisitionMetadata.objects.filter(
                entry_status=REQUIRED,
                model="edc_metadata.subjectrequisition",
                panel_name=self.panel_two.name,
                visit_code=subject_visit.visit_code,
            ).count(),
            1,
        )

    def test_get_metadata_for_subject_visit(self):
        """Asserts can get metadata for a subject and visit code."""
        subject_visit = SubjectVisit.objects.create(
            appointment=self.appointment, reason=SCHEDULED
        )
        metadata_a = []
        for metadata in subject_visit.metadata.values():
            for md in metadata:
                try:
                    metadata_a.append(f"{md.model}.{md.panel_name}")
                except AttributeError:
                    metadata_a.append(md.model)
        metadata_a.sort()
        metadata_b = [
            crf.model
            for crf in subject_visit.schedule.visits.get(subject_visit.visit_code).crfs
        ]
        metadata_b.extend(
            [
                f"{requisition.model}.{requisition.name}"
                for requisition in subject_visit.schedule.visits.get(
                    subject_visit.visit_code
                ).requisitions
            ]
        )
        metadata_b.sort()
        self.assertEqual(metadata_a, metadata_b)

    def test_metadata_inspector(self):
        subject_visit = SubjectVisit.objects.create(
            appointment=self.appointment, reason=SCHEDULED
        )
        inspector = MetaDataInspector(
            model_cls=CrfOne,
            visit_schedule_name=subject_visit.visit_schedule_name,
            schedule_name=subject_visit.schedule_name,
            visit_code=subject_visit.visit_code,
            timepoint=subject_visit.timepoint,
        )
        self.assertEqual(len(inspector.required), 1)
        self.assertEqual(len(inspector.keyed), 0)

        CrfOne.objects.create(subject_visit=subject_visit)

        inspector = MetaDataInspector(
            model_cls=CrfOne,
            visit_schedule_name=subject_visit.visit_schedule_name,
            schedule_name=subject_visit.schedule_name,
            visit_code=subject_visit.visit_code,
            timepoint=subject_visit.timepoint,
        )
        self.assertEqual(len(inspector.required), 0)
        self.assertEqual(len(inspector.keyed), 1)

    def test_crf_updates_ok(self):
        subject_visit = SubjectVisit.objects.create(
            appointment=self.appointment, reason=SCHEDULED
        )
        CrfMetadata.objects.get(
            visit_code=subject_visit.visit_code,
            model="edc_metadata.crfone",
            entry_status=REQUIRED,
        )
        metadata_updater = MetadataUpdater(
            related_visit=subject_visit,
            target_model="edc_metadata.crfone",
        )
        metadata_updater.update(entry_status=NOT_REQUIRED)
        self.assertRaises(
            ObjectDoesNotExist,
            CrfMetadata.objects.get,
            visit_code=subject_visit.visit_code,
            model="edc_metadata.crfone",
            entry_status=REQUIRED,
        )

        for visit_obj in SubjectVisit.objects.all():
            if visit_obj == subject_visit:
                try:
                    CrfMetadata.objects.get(
                        visit_code=visit_obj.visit_code,
                        model="edc_metadata.crfone",
                        entry_status=NOT_REQUIRED,
                    )
                except ObjectDoesNotExist as e:
                    self.fail(e)
            else:
                self.assertRaises(
                    ObjectDoesNotExist,
                    CrfMetadata.objects.get,
                    visit_code=visit_obj.visit_code,
                    model="edc_metadata.crfone",
                    entry_status=NOT_REQUIRED,
                )

    def test_crf_invalid_model(self):
        subject_visit = SubjectVisit.objects.create(
            appointment=self.appointment, reason=SCHEDULED
        )
        metadata_updater = MetadataUpdater(
            related_visit=subject_visit,
            target_model="edc_metadata.blah",
        )
        self.assertRaises(
            TargetModelLookupError, metadata_updater.update, entry_status=NOT_REQUIRED
        )

    def test_crf_model_not_scheduled(self):
        subject_visit = SubjectVisit.objects.create(
            appointment=self.appointment, reason=SCHEDULED
        )
        metadata_updater = MetadataUpdater(
            related_visit=subject_visit,
            target_model="edc_metadata.crfseven",
        )
        self.assertRaises(
            TargetModelNotScheduledForVisit,
            metadata_updater.update,
            entry_status=NOT_REQUIRED,
        )



from collections import OrderedDict
from copy import deepcopy

from dateutil.relativedelta import relativedelta
from django import forms
from django.conf import settings
from django.contrib.auth.models import User
from django.contrib.sites.models import Site
from django.test import TestCase, override_settings
from edc_appointment.constants import IN_PROGRESS_APPT, INCOMPLETE_APPT
from edc_appointment.models import Appointment
from edc_constants.constants import FEMALE
from edc_facility import import_holidays
from edc_reference import site_reference_configs
from edc_utils import get_utcnow
from edc_visit_schedule import site_visit_schedules
from edc_visit_schedule.constants import DAY1, MONTH1, MONTH3, MONTH6, WEEK2
from edc_visit_tracking.constants import SCHEDULED
from model_bakery import baker

from edc_metadata import (
    KEYED,
    NOT_REQUIRED,
    REQUIRED,
    TargetModelNotScheduledForVisit,
    site_metadata_rules,
)
from edc_metadata.metadata import CrfMetadataGetter

from ...metadata_rules import (
    CrfRule,
    CrfRuleGroup,
    PersistantSingletonMixin,
    PredicateCollection,
)
from ...models import CrfMetadata
from ..models import CrfOne, SubjectConsent, SubjectVisit
from ..visit_schedule2 import visit_schedule


class CrfOneForm(forms.ModelForm):
    class Meta:
        model = CrfOne
        fields = "__all__"


@override_settings(
    EDC_PROTOCOL_STUDY_OPEN_DATETIME=get_utcnow() - relativedelta(years=3),
    EDC_PROTOCOL_STUDY_CLOSE_DATETIME=get_utcnow() + relativedelta(years=3),
)
class TestPersistantSingleton(TestCase):
    @classmethod
    def setUpTestData(cls):
        import_holidays()

    def setUp(self):
        site_visit_schedules._registry = {}
        site_visit_schedules.loaded = False
        site_visit_schedules.register(visit_schedule)

        site_reference_configs.register_from_visit_schedule(
            visit_models={"edc_appointment.appointment": "edc_metadata.subjectvisit"}
        )

        site_metadata_rules.registry = OrderedDict()
        site_metadata_rules.register(self.rule_group)

        self.user = User.objects.create(username="erik")

        self.subject_identifier = "1111111"

        subject_consent = SubjectConsent.objects.create(
            subject_identifier=self.subject_identifier,
            consent_datetime=get_utcnow() - relativedelta(months=24),
        )
        _, self.schedule = site_visit_schedules.get_by_onschedule_model(
            "edc_metadata.onschedule"
        )
        self.schedule.put_on_schedule(
            subject_identifier=self.subject_identifier,
            onschedule_datetime=subject_consent.consent_datetime,
        )
        self.appointment = Appointment.objects.get(
            subject_identifier=self.subject_identifier,
            visit_code=self.schedule.visits.first.code,
        )
        self.subject_visit = SubjectVisit.objects.create(
            appointment=self.appointment,
            subject_identifier=self.subject_identifier,
            reason=SCHEDULED,
        )

        self.data = dict(
            subject_visit=self.subject_visit,
            report_datetime=self.subject_visit.report_datetime,
            f1="blah",
            f2="blah",
            f3="blah",
        )

    @staticmethod
    def get_subject_consent():
        return baker.make_recipe(
            "edc_metadata.subjectconsent",
            user_created="erikvw",
            user_modified="erikvw",
            screening_identifier="1234",
            initials="XX",
            gender=FEMALE,
            dob=get_utcnow().date() - relativedelta(years=25),
            site=Site.objects.get(id=settings.SITE_ID),
            consent_datetime=get_utcnow() - relativedelta(months=24),
        )

    def get_subject_visit(
        self,
        visit_code=None,
        visit_code_sequence=None,
        reason=None,
        appt_datetime=None,
    ):
        reason = reason or SCHEDULED
        subject_consent = self.get_subject_consent()
        options = dict(
            subject_identifier=subject_consent.subject_identifier,
            visit_code=visit_code or DAY1,
            visit_code_sequence=(
                visit_code_sequence if visit_code_sequence is not None else 0
            ),
            reason=reason,
        )
        if appt_datetime:
            options.update(appt_datetime=appt_datetime)
        appointment = self.get_appointment(**options)
        subject_visit = SubjectVisit(
            appointment=appointment,
            reason=SCHEDULED,
            report_datetime=appointment.appt_datetime,
        )
        subject_visit.save()
        subject_visit.refresh_from_db()
        return subject_visit

    @staticmethod
    def get_next_subject_visit(subject_visit):
        appointment = subject_visit.appointment
        appointment.appt_status = INCOMPLETE_APPT
        appointment.save()
        appointment.refresh_from_db()
        next_appointment = appointment.next_by_timepoint
        next_appointment.appt_status = IN_PROGRESS_APPT
        next_appointment.save()
        subject_visit = SubjectVisit(
            appointment=next_appointment,
            reason=SCHEDULED,
            report_datetime=next_appointment.appt_datetime,
            visit_code=next_appointment.visit_code,
            visit_code_sequence=next_appointment.visit_code_sequence,
        )
        subject_visit.save()
        subject_visit.refresh_from_db()
        return subject_visit

    @property
    def rule_group(self):
        class Predicates(PersistantSingletonMixin, PredicateCollection):
            app_label = "edc_metadata"
            visit_model = "edc_metadata.subjectvisit"

            def crfone_required(self, visit, **kwargs):  # noqa
                model = f"{self.app_label}.crfone"
                return self.persistant_singleton_required(
                    visit, model=model, exclude_visit_codes=[DAY1]
                )

        pc = Predicates()

        class RuleGroup(CrfRuleGroup):
            crfone = CrfRule(
                predicate=pc.crfone_required,
                consequence=REQUIRED,
                alternative=NOT_REQUIRED,
                target_models=["crfone"],
            )

            class Meta:
                app_label = "edc_metadata"
                source_model = "edc_metadata.subjectvisit"

        return RuleGroup

    def test_baseline_not_required(self):
        site_metadata_rules.registry = OrderedDict()
        site_metadata_rules.register(self.rule_group)
        form = CrfOneForm(data=self.data)
        form.is_valid()
        self.assertEqual({}, form._errors)
        self.assertRaises(TargetModelNotScheduledForVisit, form.save)

        crf_metadata_getter = CrfMetadataGetter(appointment=self.subject_visit.appointment)
        self.assertFalse(
            crf_metadata_getter.metadata_objects.filter(
                model="edc_metadata.crfone", entry_status=REQUIRED
            ).exists()
        )

    def test_1005_required(self):
        site_metadata_rules.registry = OrderedDict()
        site_metadata_rules.register(self.rule_group)
        subject_visit = self.get_next_subject_visit(self.subject_visit)
        self.assertEqual(subject_visit.visit_code, WEEK2)
        crf_metadata_getter = CrfMetadataGetter(appointment=subject_visit.appointment)
        self.assertTrue(
            crf_metadata_getter.metadata_objects.filter(model="edc_metadata.crfone").exists()
        )
        self.assertEqual(
            crf_metadata_getter.metadata_objects.get(
                model="edc_metadata.crfone", visit_code=WEEK2
            ).entry_status,
            REQUIRED,
        )

        self.assertEqual(CrfMetadata.objects.filter(model="edc_metadata.crfone").count(), 1)

        data = deepcopy(self.data)
        data.update(subject_visit=subject_visit, report_datetime=subject_visit.report_datetime)
        form = CrfOneForm(data=data)
        form.is_valid()
        self.assertEqual({}, form._errors)
        form.save()
        self.assertTrue(
            crf_metadata_getter.metadata_objects.filter(
                model="edc_metadata.crfone", entry_status=KEYED
            ).exists()
        )

    def test_visit_required_if_not_submitted(self):
        site_metadata_rules.registry = OrderedDict()
        site_metadata_rules.register(self.rule_group)
        subject_visit = self.get_next_subject_visit(self.subject_visit)
        self.assertEqual(subject_visit.visit_code, WEEK2)
        self.assertEqual(CrfMetadata.objects.filter(model="edc_metadata.crfone").count(), 1)
        self.assertEqual(
            [(WEEK2, REQUIRED)],
            [
                (obj.visit_code, obj.entry_status)
                for obj in CrfMetadata.objects.filter(model="edc_metadata.crfone").order_by(
                    "timepoint"
                )
            ],
        )

        subject_visit = self.get_next_subject_visit(subject_visit)
        self.assertEqual(subject_visit.visit_code, MONTH1)
        self.assertEqual(CrfMetadata.objects.filter(model="edc_metadata.crfone").count(), 2)
        self.assertEqual(
            [(WEEK2, NOT_REQUIRED), (MONTH1, REQUIRED)],
            [
                (obj.visit_code, obj.entry_status)
                for obj in CrfMetadata.objects.filter(model="edc_metadata.crfone").order_by(
                    "timepoint"
                )
            ],
        )

        subject_visit = self.get_next_subject_visit(subject_visit)
        self.assertEqual(subject_visit.visit_code, MONTH3)
        self.assertEqual(CrfMetadata.objects.filter(model="edc_metadata.crfone").count(), 3)
        self.assertEqual(
            [(WEEK2, NOT_REQUIRED), (MONTH1, NOT_REQUIRED), (MONTH3, REQUIRED)],
            [
                (obj.visit_code, obj.entry_status)
                for obj in CrfMetadata.objects.filter(model="edc_metadata.crfone").order_by(
                    "timepoint"
                )
            ],
        )

        subject_visit = self.get_next_subject_visit(subject_visit)
        self.assertEqual(subject_visit.visit_code, MONTH6)
        self.assertEqual(CrfMetadata.objects.filter(model="edc_metadata.crfone").count(), 4)
        self.assertEqual(
            [
                (WEEK2, NOT_REQUIRED),
                (MONTH1, NOT_REQUIRED),
                (MONTH3, NOT_REQUIRED),
                (MONTH6, REQUIRED),
            ],
            [
                (obj.visit_code, obj.entry_status)
                for obj in CrfMetadata.objects.filter(model="edc_metadata.crfone").order_by(
                    "timepoint"
                )
            ],
        )

    def test_1010_required_if_not_submitted(self):
        site_metadata_rules.registry = OrderedDict()
        site_metadata_rules.register(self.rule_group)
        subject_visit = self.get_next_subject_visit(self.subject_visit)
        subject_visit = self.get_next_subject_visit(subject_visit)
        self.assertEqual(subject_visit.visit_code, MONTH1)
        crf_metadata_getter = CrfMetadataGetter(appointment=subject_visit.appointment)
        self.assertTrue(
            crf_metadata_getter.metadata_objects.filter(model="edc_metadata.crfone").exists()
        )
        self.assertEqual(
            crf_metadata_getter.metadata_objects.get(
                model="edc_metadata.crfone", visit_code=MONTH1
            ).entry_status,
            REQUIRED,
        )
        data = deepcopy(self.data)
        data.update(subject_visit=subject_visit, report_datetime=subject_visit.report_datetime)
        form = CrfOneForm(data=data)
        form.is_valid()
        self.assertEqual({}, form._errors)
        form.save()
        self.assertTrue(
            crf_metadata_getter.metadata_objects.filter(
                model="edc_metadata.crfone", entry_status=KEYED
            ).exists()
        )

    def test_1030_required_if_not_submitted(self):
        site_metadata_rules.registry = OrderedDict()
        site_metadata_rules.register(self.rule_group)
        subject_visit = self.get_next_subject_visit(self.subject_visit)
        subject_visit = self.get_next_subject_visit(subject_visit)
        subject_visit = self.get_next_subject_visit(subject_visit)
        self.assertEqual(subject_visit.visit_code, MONTH3)
        crf_metadata_getter = CrfMetadataGetter(appointment=subject_visit.appointment)
        self.assertTrue(
            crf_metadata_getter.metadata_objects.filter(model="edc_metadata.crfone").exists()
        )
        self.assertEqual(
            crf_metadata_getter.metadata_objects.get(
                model="edc_metadata.crfone", visit_code=MONTH3
            ).entry_status,
            REQUIRED,
        )
        data = deepcopy(self.data)
        data.update(subject_visit=subject_visit, report_datetime=subject_visit.report_datetime)
        form = CrfOneForm(data=data)
        form.is_valid()
        self.assertEqual({}, form._errors)
        form.save()
        self.assertTrue(
            crf_metadata_getter.metadata_objects.filter(
                model="edc_metadata.crfone", entry_status=KEYED
            ).exists()
        )

    def test_1030_not_required_if_submitted(self):
        site_metadata_rules.registry = OrderedDict()
        site_metadata_rules.register(self.rule_group)
        subject_visit_1005 = self.get_next_subject_visit(self.subject_visit)
        subject_visit_1010 = self.get_next_subject_visit(subject_visit_1005)
        subject_visit_1030 = self.get_next_subject_visit(subject_visit_1010)
        self.assertEqual(subject_visit_1030.visit_code, MONTH3)
        crf_metadata_getter = CrfMetadataGetter(appointment=subject_visit_1030.appointment)
        self.assertTrue(
            crf_metadata_getter.metadata_objects.filter(
                model="edc_metadata.crfone", entry_status=REQUIRED
            ).exists()
        )
        data = deepcopy(self.data)
        data.update(
            subject_visit=subject_visit_1010,
            report_datetime=subject_visit_1010.report_datetime,
        )
        form = CrfOneForm(data=data)
        form.is_valid()
        self.assertEqual({}, form._errors)
        form.save()
        crf_metadata_getter = CrfMetadataGetter(appointment=subject_visit_1005.appointment)
        self.assertTrue(
            crf_metadata_getter.metadata_objects.filter(
                model="edc_metadata.crfone", entry_status=NOT_REQUIRED
            ).exists()
        )
        crf_metadata_getter = CrfMetadataGetter(appointment=subject_visit_1010.appointment)
        self.assertTrue(
            crf_metadata_getter.metadata_objects.filter(
                model="edc_metadata.crfone", entry_status=KEYED
            ).exists()
        )
        crf_metadata_getter = CrfMetadataGetter(appointment=subject_visit_1030.appointment)
        self.assertTrue(
            crf_metadata_getter.metadata_objects.filter(
                model="edc_metadata.crfone", entry_status=NOT_REQUIRED
            ).exists()
        )



from __future__ import annotations

from typing import TYPE_CHECKING

from django.apps import apps as django_apps
from edc_reference import site_reference_configs
from edc_visit_tracking.constants import MISSED_VISIT

from edc_metadata.constants import CRF
from edc_metadata.metadata_handler import MetadataHandler

if TYPE_CHECKING:
    from edc_visit_tracking.model_mixins import VisitModelMixin


class TargetModelNotScheduledForVisit(Exception):
    pass


class TargetModelConflict(Exception):
    pass


class TargetModelMissingManagerMethod(Exception):
    pass


class TargetModelLookupError(Exception):
    pass


class TargetModelMissingMetadata(Exception):
    pass


class TargetHandler:

    """A class that gets the target model "model instance"
    for a given visit, if it exists.

    If target model is not scheduled for this visit a
    TargetModelNotScheduledForVisit exception will be raised.
    """

    metadata_handler_cls = MetadataHandler
    metadata_category = CRF
    metadata_model = "edc_metadata.crfmetadata"

    def __init__(self, model: str = None, related_visit: VisitModelMixin = None):

        self.model = model
        self.related_visit = related_visit  # visit model instance
        self.metadata_model_cls = django_apps.get_model(
            self.metadata_model
        )  # .get_metadata_model(self.metadata_category)
        if self.model == self.related_visit._meta.label_lower:
            raise TargetModelConflict(
                f"Target model and visit model are the same! "
                f"Got {self.model}=={self.related_visit._meta.label_lower}"
            )

        try:
            django_apps.get_model(self.model)
        except LookupError as e:
            raise TargetModelLookupError(
                f"{self.metadata_category} target model name is invalid. Got {e}"
            )
        if self.related_visit.reason != MISSED_VISIT:
            self.raise_on_not_scheduled_for_visit()
        self.metadata_obj = self.metadata_handler.metadata_obj
        if not self.metadata_obj and related_visit.reason != MISSED_VISIT:
            raise TargetModelMissingMetadata(
                f"{self.metadata_model} model instance unexpectedly does not exist! "
                f"Got model: `{model}` visit: `{related_visit}`. "
                f"visit reason={related_visit.reason}`."
            )

    def __repr__(self):
        return (
            f"<{self.__class__.__name__}({self.model}, {self.related_visit}), "
            f"{self.metadata_model_cls._meta.label_lower}>"
        )

    @property
    def reference_model_cls(self):
        reference_model = site_reference_configs.get_reference_model(name=self.model)
        return django_apps.get_model(reference_model)

    @property
    def object(self):
        """Returns a reference model instance for the "target".

        Recall the CRF/Requisition is not queried directly but rather
        represented by a model instance from edc_reference.
        """
        return self.reference_model_cls.objects.filter_crf_for_visit(
            name=self.model, visit=self.related_visit
        )

    @property
    def metadata_handler(self):
        return self.metadata_handler_cls(
            metadata_model=self.metadata_model,
            model=self.model,
            related_visit=self.related_visit,
        )

    @property
    def models(self):
        """Returns a list of models for this visit."""
        if self.related_visit.visit_code_sequence != 0:
            forms = (
                self.related_visit.visit.unscheduled_forms + self.related_visit.visit.prn_forms
            )
        elif self.related_visit.reason == MISSED_VISIT:
            forms = self.related_visit.visit.missed_forms
        else:
            forms = self.related_visit.visit.forms + self.related_visit.visit.prn_forms
        return list(set([form.model for form in forms]))

    def raise_on_not_scheduled_for_visit(self):
        """Raises an exception if model is not scheduled
        for this visit.

        PRN forms are added to the list of scheduled forms
        for the conditional eval.
        """
        if self.model not in self.models:
            raise TargetModelNotScheduledForVisit(
                f"Target model `{self.model}` is not scheduled "
                f"(nor a PRN) for visit '{self.related_visit.visit_code}."
                f"{self.related_visit.visit_code_sequence}' "
                f"subject '{self.related_visit.subject_identifier}'."
            )
